from typing import Optional
from math import floor, sqrt
from spacy.language import Language
from spacy.tokens import Doc
from statistics import commonlists

@Language.factory("readability")
def get_readability(nlp: Language, name: str):
    return Readability(nlp, name)

class Readability:
    def __init__(self, nlp: Language, name: str = "readability"):

        """
        nlp: an instance of spacy
        """
        self.nlp = nlp
        self.name = "readability"
        self.common = commonlists.DC_3000


    def flesch_reading_ease(self, counts, scores):
        return round(206.835 - (1.015 * scores["ASL"]) - (84.6 * scores["ASW"]),2)

    def flesch_kincaid_grade(self, counts, scores):
        return round((0.39 * scores["ASL"]) + (11.8 * scores["ASW"]) - 15.59,2)

    def automated_readability_index(self, counts, scores):
        return round((4.71 * scores["AWL"]) + (0.5 * scores["ASL"]) - 21.43,2)

    def smog_index(self, counts, scores):
        return round(1.0430 * sqrt(counts["poly"] * (30 / counts["sentences"])) + 3.1291,2)

    def dale_chall(self, counts, scores):
        base = 3.6365 if scores["uncommon_index"] > 5 else 0
        return round(base + (0.1579 * scores["uncommon_index"]) + (0.0496 * scores["ASL"]), 2)

    def dale_chall_grade(self, counts, scores):
        return 4 + ((floor(scores["dale_chall"]) - 4) * 2)

    def get_uncommon(self, doc):
        return [t.text for t in doc if t.is_alpha and t.pos_ != "PROPN" and t.lower_ not in self.common]

    def get_counts(self, doc):
        words = [t for t in doc if t.is_alpha]
        return {
            "sentences": len([s for s in doc.sents if len(s.text.strip()) > 0]),
            "words": len(words),
            "stop": len([w for w in doc if w.is_stop]),
            "syllables": sum([w._.syllables_count for w in words]),
            "characters": sum([len(w.text.strip()) for w in words]),
            "poly": len([w for w in words if w._.syllables_count > 2]),
            "uncommon": len(doc._.uncommon),
            #"passive": len(doc.spans["passive"])
        }


    def get_scores(self, counts):
        scores = {}

        scores["ASL"] = 0
        scores["ASW"] = 0
        scores["AWL"] = 0
        scores["flesch_reading_ease"] = 0
        scores["flesch_kincaid_grade"] = 0
        scores["automated_readability_index"] = 0
        scores["smog_index"] = 0
        #scores["passive_index"] = 0
        scores["uncommon_index"] = 0
        scores["poly_index"] = 0
        scores["dale_chall"] = 0
        scores["dale_chall_grade"] = 0

        if counts["words"] > 0 and counts["sentences"] > 0:
            scores["ASL"] = round(counts["words"] / counts["sentences"], 2)
            scores["AWL"] = round(counts["characters"] / counts["words"], 2)
            scores["ASW"] = round(counts["syllables"] / counts["words"], 2)
            #scores["passive_index"] = round((counts["passive"]/counts["sentences"] * 100),2)
            scores["poly_index"] = round((counts["poly"]/counts["words"] * 100),2)
            scores["uncommon_index"] = round((counts["uncommon"]/counts["words"] * 100),2)
            scores["flesch_reading_ease"] = self.flesch_reading_ease(counts, scores)
            scores["flesch_kincaid_grade"] = self.flesch_kincaid_grade(counts, scores)
            scores["automated_readability_index"] = self.automated_readability_index(counts, scores)
            scores["smog_index"] = self.smog_index(counts, scores)
            scores["dale_chall"] = self.dale_chall(counts, scores)
            scores["dale_chall_grade"] = self.dale_chall_grade(counts, scores)

        return scores


    def __call__(self, doc):
        # Add the extension to Doc
        Doc.set_extension("readability", default=None, force=True)
        Doc.set_extension("counts", default=None, force=True)
        Doc.set_extension("uncommon", default=None, force=True)

        doc._.uncommon = self.get_uncommon(doc)
        doc._.counts = self.get_counts(doc)
        doc._.readability = self.get_scores(doc._.counts)

        doc.cats = []
        for score in doc._.readability:
            doc.cats.append({"label":score,"score":doc._.readability[score]})

        for score in doc._.counts:
            doc.cats.append({"label":score,"score":doc._.counts[score]})

        doc.cats.append({"label":"formality","score":doc._.formality})

        return doc
